#
#----------------------------------------------------------------------
#   GCodePlot2D
#----------------------------------------------------------------------
#import time as TIM
#import random as RND
import numpy as NUM
from enum import Enum
#
import Define as DEF
import Helper as HLP
#import Task as TSK
#import Command as CMD
#
#---------------------------------------------------------
# Type - EStateInterface
#---------------------------------------------------------
class EStateInterface(Enum):
    Idle = 0
    TransmitCommand = 1
    ReceiveCommand = 2
#---------------------------------------------------------
# Type - CGCodeSimulator
#---------------------------------------------------------
class CGCodeSimulator():
    #
    def __init__(self):
        self.State = 0 # Idle EStateInterface.Idle
        # WorkingParameter
        self.UnitMM = True
        # WorkingPlane
        self.PlaneXY = True
        self.PlaneYZ = False
        self.PlaneZX = False
        self.X0 = float(0.0) # [mm] - PresetX
        self.X1 = float(0.1) # [mm]
        self.Y0 = float(0.0) # [mm] - PresetY
        self.Y1 = float(0.1) # [mm]
        self.Z0 = float(0.0) # [mm] - PresetZ
        self.Z1 = float(0.1) # [mm]
        # reset:
        self.I = float(0.0) # [mm] # I-Parameter Arc
        self.J = float(0.0) # [mm] # J-Parameter Arc
        self.R = float(0.0) # [mm] # Radius == 0 : IJ->R
        #
        self.E =  float(0.0) # [mm]
        self.FL = DEF.FEED_LOW  # [mm/s] - FeedLow (G1,G2,G3...)
        self.FH = DEF.FEED_HIGH # [mm/s] - FeedHigh (G0)
        # Plot
        self.VX0 = [] # PlotVectorCollected-X
        self.VY0 = [] # PlotVectorCollected-Y
        self.VZ0 = [] # PlotVectorCollected-Z
        self.VX1 = [] # PlotVectorActual-X
        self.VY1 = [] # PlotVectorActual-Y
        self.VZ1 = [] # PlotVectorActual-Z
        self.XL =  0.0
        self.XH = 10.0
        self.InitX = True
        self.YL =  0.0
        self.YH = 10.0
        self.InitY = True
        self.ZL =  0.0
        self.ZH = 10.0
        self.InitZ = True
        self.LP = 0.0
        self.TP = 0.0
        # GCode
        self.GCode = None
        return
    #--------------------------------------------------------------------
    #   Property
    #--------------------------------------------------------------------
    def IsBusy(self):
        return (0 != self.State)
    #
    def GetVX0(self):
        return self.VX0
    def GetVY0(self):
        return self.VY0
    def GetVX1(self):
        return self.VX1
    def GetVY1(self):
        return self.VY1
    def GetXL(self):
        return self.XL
    def GetXH(self):
        return self.XH
    def GetYL(self):
        return self.YL
    def GetYH(self):
        return self.YH
    def GetTP(self):
        return self.TP


    def CalculateWAWBR(self):
        D2 = NUM.square(self.X1 - self.X0) + NUM.square(self.Y1 - self.Y0)
        if (0.0 == self.R): # IJ->R
            XM = float(self.X0 + self.I)
            YM = float(self.Y0 + self.J)
            XD = self.X0 + (self.X1 - self.X0) / 2
            YD = self.Y0 + (self.Y1 - self.Y0) / 2
            WA = HLP.ArcTanPos(self.Y0 - YM, self.X0 - XM)
            WB = HLP.ArcTanPos(self.Y1 - YM, self.X1 - XM)
            H2 = NUM.square(XD - XM) + NUM.square(YD - YM)
            R = NUM.sqrt(D2 / 4.0 + H2)
        else: # R->IJ
            R = self.R
            H = NUM.sqrt(NUM.square(self.R) - D2 / 4.0)
            XD = self.X0 + (self.X1 - self.X0) / 2.0
            YD = self.Y0 + (self.Y1 - self.Y0) / 2.0
            WG = HLP.ArcTanPos(self.Y1 - self.Y0, self.X1 - self.X0)
            XM = XD - H * NUM.sin(WG)
            YM = YD + H * NUM.cos(WG)
            WA = HLP.ArcTanPos(self.Y0 - YM, self.X0 - XM)
            WB = HLP.ArcTanPos(self.Y1 - YM, self.X1 - XM)
        return [WA, WB, R]

    #
    # F = L / T -> T = D / F
    #
    def BuildGCodeData(self, code):
#----------------------------------------------------------------------
#   G0 - G00
#----------------------------------------------------------------------
        if ('G00' == code) or ('G0' == code):
            D = NUM.sqrt(NUM.square(self.X1 - self.X0) + NUM.square(self.Y1 - self.Y0))
            self.LP = D
            self.TP = self.LP / self.FH
            print('G0-F[{0}]mm/s-L[{1}]mm-T[{2}]s'.format(self.FH, self.LP, self.TP))
            NP = int(0.5 + self.TP / DEF.TIME_REFRESH)
            if (NP <= 2):
                self.VX1.append(self.X0)
                self.VY1.append(self.Y0)
                self.VZ1.append(self.Z0)
                self.VX1.append(self.X1)
                self.VY1.append(self.Y1)
                self.VZ1.append(self.Z1)
            else:
                DX = (self.X1 - self.X0) / NP
                DY = (self.Y1 - self.Y0) / NP
                for PI in range(0, 1 + NP):
                    XI = self.X0 + DX * PI
                    YI = self.Y0 + DY * PI
                    self.VX1.append(XI)
                    self.VY1.append(YI)
                    self.VZ1.append(self.Z1)
            self.X0 = self.X1
            self.Y0 = self.Y1
            self.Z0 = self.Z1
            return ' ; Rapid Positioning'
#----------------------------------------------------------------------
#   G1 - G01
#----------------------------------------------------------------------
        if ('G01' == code) or ('G1' == code):
            D = NUM.sqrt(NUM.square(self.X1 - self.X0) + NUM.square(self.Y1 - self.Y0))
            self.LP = D
            self.TP = self.LP / self.FL
            print('G1-F[{0}]mm/s-L[{1}]mm-T[{2}]s'.format(self.FL, self.LP, self.TP))
            NP = int(0.5 + self.TP / DEF.TIME_REFRESH)
            if (NP <= 2):
                self.VX1.append(self.X0)
                self.VY1.append(self.Y0)
                self.VZ1.append(self.Z0)
                self.VX1.append(self.X1)
                self.VY1.append(self.Y1)
                self.VZ1.append(self.Z1)
            else:
                DX = (self.X1 - self.X0) / NP
                DY = (self.Y1 - self.Y0) / NP
                for PI in range(0, 1 + NP):
                    XI = self.X0 + DX * PI
                    YI = self.Y0 + DY * PI
                    self.VX1.append(XI)
                    self.VY1.append(YI)
                    self.VZ1.append(self.Z1)
            self.X0 = self.X1
            self.Y0 = self.Y1
            self.Z0 = self.Z1
            return ' ; Cut Line Moving'
#----------------------------------------------------------------------
#   G2 - G02
#----------------------------------------------------------------------
        if ('G02' == code) or ('G2' == code):
            D2 = NUM.square(self.X1 - self.X0) + NUM.square(self.Y1 - self.Y0) # later: Arc!!!
            D = NUM.sqrt(D2)
            [WA, WB, R] = self.CalculateWAWBR()
            self.LP = abs(R * (WB - WA))
            self.TP = self.LP / self.FL
            print('G2-F[{0}]mm/s-L[{1}]mm-T[{2}]s'.format(self.FL, self.LP, self.TP))
            if (0.0 == self.R): # IJ->R
                # debug print('~~~> G02 - IJ')
                XM = float(self.X0 + self.I)
                YM = float(self.Y0 + self.J)
                XD = self.X0 + (self.X1 - self.X0) / 2
                YD = self.Y0 + (self.Y1 - self.Y0) / 2
                WA = HLP.ArcTanPos(self.Y0 - YM, self.X0 - XM)
                WB = HLP.ArcTanPos(self.Y1 - YM, self.X1 - XM)
                DAB = HLP.PIDRAD - WB + WA
                NP = int(0.5 + self.TP / DEF.TIME_REFRESH)
                if (NP <= 2):
                    self.VX1.append(self.X0)
                    self.VY1.append(self.Y0)
                    self.VZ1.append(self.Z0)
                    self.VX1.append(self.X1)
                    self.VY1.append(self.Y1)
                    self.VZ1.append(self.Z1)
                else:
                    H2 = NUM.square(XD - XM) + NUM.square(YD - YM)
                    R = NUM.sqrt(D2 / 4.0 + H2)
                    for PI in range(0, 1 + NP):
                        SCA = -DAB * PI / NP + WA
                        XI = XM + R * NUM.cos(SCA)
                        YI = YM + R * NUM.sin(SCA)
                        self.VX1.append(XI)
                        self.VY1.append(YI)
                        self.VZ1.append(self.Z1)
            else: # R->IJ
                # debug print('~~~> G02 - R')
                H = NUM.sqrt(NUM.square(self.R) - D2 / 4.0)
                XD = self.X0 + (self.X1 - self.X0) / 2.0
                YD = self.Y0 + (self.Y1 - self.Y0) / 2.0
                WG = HLP.ArcTanPos(self.Y1 - self.Y0, self.X1 - self.X0)
                XM = XD - H * NUM.sin(WG)
                YM = YD + H * NUM.cos(WG)
                WA = HLP.ArcTanPos(self.Y0 - YM, self.X0 - XM)
                WB = HLP.ArcTanPos(self.Y1 - YM, self.X1 - XM)
                DAB = HLP.PIDRAD - WB + WA
                NP = int(0.5 + self.TP / DEF.TIME_REFRESH)
                if (NP <= 2):
                    self.VX1.append(self.X0)
                    self.VY1.append(self.Y0)
                    self.VZ1.append(self.Z0)
                    self.VX1.append(self.X1)
                    self.VY1.append(self.Y1)
                    self.VZ1.append(self.Z1)
                else:
                    for PI in range(0, 1 + NP):
                        SCA = -DAB * PI / NP + WA
                        XI = XM + self.R * NUM.cos(SCA)
                        YI = YM + self.R * NUM.sin(SCA)
                        self.VX1.append(XI)
                        self.VY1.append(YI)
                        self.VZ1.append(self.Z1)
            # reset:
            self.I = 0.0
            self.J = 0.0
            self.R = 0.0
            self.X0 = self.X1
            self.Y0 = self.Y1
            self.Z0 = self.Z1
            return ' ; Cut Arc Clockwise'
#----------------------------------------------------------------------
#   G3 - G03
#----------------------------------------------------------------------
        if ('G03' == code) or ('G3' == code):
            # debug print('~~~>>> G03 <<<')
            D2 = NUM.square(self.X1 - self.X0) + NUM.square(self.Y1 - self.Y0) # later: Arc!!!
            D = NUM.sqrt(D2)
            [WA, WB, R] = self.CalculateWAWBR()
            self.LP = abs(R * (WB - WA))
            self.TP = self.LP / self.FL
            print('G3-F[{0}]mm/s-L[{1}]mm-T[{2}]s'.format(self.FL, self.L, self.TP))
            if (0.0 == self.R): # IJ->R
                XM = float(self.X0 + self.I)
                YM = float(self.Y0 + self.J)
                XD = self.X0 + (self.X1 - self.X0) / 2
                YD = self.Y0 + (self.Y1 - self.Y0) / 2
                WA = HLP.ArcTanPos(self.Y0 - YM, self.X0 - XM)
                WB = HLP.ArcTanPos(self.Y1 - YM, self.X1 - XM)
                H2 = NUM.square(XD - XM) + NUM.square(YD - YM)
                R = NUM.sqrt(D2 / 4.0 + H2)
                #
                NP = int(0.5 + self.TP / DEF.TIME_REFRESH)
                if (NP <= 2):
                    self.VX1.append(self.X0)
                    self.VY1.append(self.Y0)
                    self.VZ1.append(self.Z0)
                    self.VX1.append(self.X1)
                    self.VY1.append(self.Y1)
                    self.VZ1.append(self.Z1)
                else:
                    for PI in range(0, 1 + NP):
                        SCA = (WB - WA) * PI / NP + WA
                        XI = XM + R * NUM.cos(SCA)
                        YI = YM + R * NUM.sin(SCA)
                        self.VX1.append(XI)
                        self.VY1.append(YI)
                        self.VZ1.append(self.Z1)
            else: # R->IJ
                # debug print('~~~> G03 - R')
                H = NUM.sqrt(NUM.square(self.R) - D2 / 4.0)
                XD = self.X0 + (self.X1 - self.X0) / 2.0
                YD = self.Y0 + (self.Y1 - self.Y0) / 2.0
                WG = HLP.ArcTanPos(self.Y1 - self.Y0, self.X1 - self.X0)
                XM = XD - H * NUM.sin(WG)
                YM = YD + H * NUM.cos(WG)
                WA = HLP.ArcTanPos(self.Y0 - YM, self.X0 - XM)
                WB = HLP.ArcTanPos(self.Y1 - YM, self.X1 - XM)
                #
                NP = int(0.5 + self.TP / DEF.TIME_REFRESH)
                if (NP <= 2):
                    self.VX1.append(self.X0)
                    self.VY1.append(self.Y0)
                    self.VZ1.append(self.Z0)
                    self.VX1.append(self.X1)
                    self.VY1.append(self.Y1)
                    self.VZ1.append(self.Z1)
                else:
                    for PI in range(0, 1 + NP):
                        SCA = (WB - WA) * PI / NP + WA
                        XI = XM + self.R * NUM.cos(SCA)
                        YI = YM + self.R * NUM.sin(SCA)
                        self.VX1.append(XI)
                        self.VY1.append(YI)
                        self.VZ1.append(self.Z1)
            # reset:
            self.I = 0.0
            self.J = 0.0
            self.R = 0.0
            self.X0 = self.X1
            self.Y0 = self.Y1
            self.Z0 = self.Z1
            return ' ; Cut Arc Counter-Clockwise'
#----------------------------------------------------------------------
#   G17 ...
#----------------------------------------------------------------------
        if ('G17' == code):
            return ' ; Arc in XY-Plane'
        if ('G18' == code):
            return ' ; Arc in XZ-Plane'
        if ('G19' == code):
            return ' ; Arc in YZ-Plane'
        if ('G20' == code):
            return ' ; Unit Inches'
        if ('G21' == code):
            return ' ; Unit Millimeters'
        if ('G40' == code):
            return ' ; Cutter Compensation Off'
        if ('G49' == code):
            return ' ; Cancel Tool Length Offset'
        if ('G53' == code):
            return ' ; Use Machine Coordinates'
        if ('G54' == code):
            return ' ; Activate Saved Origin'
        if ('G80' == code):
            return ' ; Cancel'
        if ('G90' == code):
            return ' ; Mode Absolute'
        if ('G91' == code):
            return ' ; Mode Absolute'
        return ' '
    #
    def BuildMCodeData(self, code):
        if ('M0' == code) or ('M00' == code):
            return ' ; Pause'
        if ('M1' == code) or ('M01' == code):
            return ' ; Pause on StopSwitch'
        if ('M2' == code) or ('M02' == code):
            return ' ; Program End'
        if ('M3' == code) or ('M03' == code):
            return ' ; Start Spindle Clockwise'
        if ('M4' == code) or ('M04' == code):
            return ' ; Start Spindle CounterClockwise'
        if ('M5' == code) or ('M05' == code):
            return ' ; Stop Spindle'
        if ('M6' == code) or ('M06' == code):
            return ' ; Select Tool'
        if ('M8' == code):
            return ' ; Coolant On'
        if ('M9' == code):
            return ' ; Coolant Off'
        if ('M25' == code):
            return ' ; Pause SDCard Execution'
        if ('M30' == code):
            return ' ; Program End'
        if ('M99' == code):
            return ' ; Return from SubProgram'
        return ' '
    #
    def RefreshExtremaX(self, valuex):
        if (self.InitX):
            self.InitX = False
            self.XL = self.XL - self.XL / 10.0
            self.XH = valuex
        else:
            if (valuex < self.XL):
                self.XL = valuex
                return
            if (self.XH < valuex):
                self.XH = valuex
                return
    def RefreshExtremaY(self, valuey):
        if (self.InitY):
            self.InitY = False
            self.YL = self.YL - self.YL / 10.0
            self.YH = valuey
        else:
            if (valuey < self.YL):
                self.YL = valuey
                return
            if (self.YH < valuey):
                self.YH = valuey
                return
    def RefreshExtremaZ(self, valuez):
        if (self.InitZ):
            self.InitZ = False
            self.ZL = self.ZL - self.ZL / 10.0
            self.ZH = valuez
        else:
            if (valuez < self.ZL):
                self.ZL = valuez
                return
            if (self.ZH < valuez):
                self.ZH = valuez
                return
    #
    def BuildGCodeParameter(self, gcode):
        self.GCode = gcode.Code # string
        for Parameter in gcode.Parameterlist:
            if ('X' == Parameter.Code): # float
                self.X1 = float(Parameter.Value)
                self.RefreshExtremaX(self.X1)
            elif ('Y' == Parameter.Code): # float
                self.Y1 = float(Parameter.Value)
                self.RefreshExtremaY(self.Y1)
            elif ('Z' == Parameter.Code): # float
                self.Z1 = float(Parameter.Value)
                self.RefreshExtremaZ(self.Z1)
            elif ('E' == Parameter.Code): # float
                self.E = float(Parameter.Value)
            elif ('F' == Parameter.Code): # float
                self.F = float(Parameter.Value)
            elif ('I' == Parameter.Code): # float
                self.I = float(Parameter.Value)
                #debug print('!!!!!!!!!!!!!!!!!!!! I=', self.I)
            elif ('J' == Parameter.Code): # float
                self.J = float(Parameter.Value)
                #debug print('!!!!!!!!!!!!!!!!!!!! J=', self.J)
            elif ('R' == Parameter.Code): # float
                self.R = float(Parameter.Value)
                #debug
                print('!!!!!!!!!!!!!!!!!!!! R=', self.R)
            # elif ('H' == Parameter.Code): # float
            #     self.H = float(Parameter.Value)
            # elif ('S' == Parameter.Code): # float
            #     self.S = float(Parameter.Value)
    #
    def ExecuteGCode(self, gcode):
        ### print('Type[' + str(type(gcode)) + ']')
        self.State = 1
        if (0 < len(self.VX1)):
            self.VX0.extend(self.VX1)
            self.VY0.extend(self.VY1)
            self.VZ0.extend(self.VZ1)
            self.VX1 = []
            self.VY1 = []
            self.VZ1 = []
        self.BuildGCodeParameter(gcode)
        ReportText = ''
        if ('<class \'Command.CCommand\'>' == str(type(gcode))):
            # NC? ReportText += 'CComment[' + gcode.Comment + ']'
            # nothing do do...
            self.State = 0
            return [True, ReportText]
        elif ('<class \'Command.CGCommand\'>' == str(type(gcode))):
            self.BuildGCodeData(gcode.Code)
            self.State = 0
            return [True, ReportText]
        elif ('<class \'Command.CMCommand\'>' == str(type(gcode))):
            self.BuildMCodeData(gcode.Code)
            self.State = 0
            return [True, ReportText]
        else:
            self.State = 0
            return [False, 'Invalid GCode']
    #
    #
#
#
#################################################################################
